/*
	CA-driver for TwinHan DST Frontend

	Copyright (C) 2004 Manu Abraham

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/



#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/string.h>

#include <linux/dvb/ca.h>
#include "dvbdev.h"
#include "dvb_frontend.h"

#include "dst_ca.h"
#include "dst_common.h"



static unsigned int dst_ca_verbose = 0;
MODULE_PARM(dst_ca_verbose, "i");
MODULE_PARM_DESC(dst_ca_verbose, "verbose startup messages, default is 1 (yes)");
static unsigned int dst_ca_debug = 0;
MODULE_PARM(dst_ca_debug, "i");
MODULE_PARM_DESC(dst_ca_debug, "debug messages, default is 0 (no)");
static unsigned int session = 0;
MODULE_PARM(session, "i");
MODULE_PARM_DESC(session, "certain hardware might have only a single session, default 0");
#define dprintk	if (dst_ca_debug) printk


/*	Module functions		*/
static int dst_ca_open (struct inode *inode, struct file *file);
static int dst_ca_release (struct inode *inode, struct file *file);
static int dst_ca_read (struct file *file, char __user * buffer, size_t length, loff_t * offset);
static int dst_ca_write (struct file *file, const char __user * buffer, size_t length, loff_t * offset);
static int dst_ca_ioctl (struct inode *inode, struct file *file, unsigned int cmd, void *arg);

/*	DVB API style implementation	*/
static int ca_send_message (struct dst_state *state, struct ca_msg *p_ca_message, void *arg);
static int ca_get_slot_descr (struct dst_state *state, struct ca_msg *p_ca_message, void *arg);
static int ca_get_slot_info (struct dst_state *state, struct ca_slot_info *p_ca_slot_info, void *arg);
static int ca_get_message (struct dst_state *state, struct ca_msg *p_ca_message, void *arg);
static int ca_get_slot_caps (struct dst_state *state, struct ca_caps *p_ca_caps, void *arg);
static int ca_set_pid (void);
static int ca_set_slot_descr (void);

/*	Device specific functions	*/
static int ca_set_pmt (struct dst_state *state, struct ca_msg *p_ca_message, struct ca_msg *hw_buffer, u8 reply, u8 query);
static int dst_check_ca_pmt (struct dst_state *state, struct ca_msg *p_ca_message, struct ca_msg *hw_buffer);


/*	Helper functions	*/
static int dst_put_ci (struct dst_state *state, u8 *data, int len, u8 *ca_string, int read);
static int do_checksum (struct ca_msg *hw_buffer);
static int message_display (int retval);



static struct file_operations dst_ca_fops = {
	.owner = THIS_MODULE,
	.ioctl = (void *)dst_ca_ioctl,
	.open = dst_ca_open,
	.release = dst_ca_release,
	.read = dst_ca_read,
	.write = dst_ca_write
};


static struct dvb_device dvbdev_ca = {
	.priv = NULL,
	.users = 1,
	.readers = 1,
	.writers = 1,
	.fops = &dst_ca_fops
};



static int dst_ca_ioctl (struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
	int retval;
	
	struct dvb_device* dvbdev = (struct dvb_device*) file->private_data;
	struct dst_state* state = (struct dst_state*) dvbdev->priv;
	struct ca_slot_info *p_ca_slot_info;
	struct ca_caps *p_ca_caps;
	struct ca_msg *p_ca_message;    

	if ((p_ca_message = (struct ca_msg *) kmalloc(sizeof (struct ca_msg), GFP_KERNEL)) == NULL) {
		printk ("%s: Memory allocation failure\n", __FUNCTION__);
		return -ENOMEM;
	}
	
	if ((p_ca_slot_info = (struct ca_slot_info *) kmalloc (sizeof (struct ca_slot_info), GFP_KERNEL)) == NULL) {
		printk ("%s: Memory allocation failure\n", __FUNCTION__);
		return -ENOMEM;
	}
	
	if ((p_ca_caps = (struct ca_caps *) kmalloc (sizeof (struct ca_caps), GFP_KERNEL)) == NULL) {
		printk ("%s: Memory allocation failure\n", __FUNCTION__);
		return -ENOMEM;
	}


	/*	We have now only the standard ioctl's, the driver is upposed to handle internals.*/
	switch (cmd) {
		case CA_GET_MSG:
			if (dst_ca_verbose) {
				dprintk ("%s: Getting message\n", __FUNCTION__);
			}
			retval = ca_get_message (state, p_ca_message, arg);
			message_display(retval);

			break;		
	    
		case CA_SEND_MSG:
			if (dst_ca_verbose) {
				dprintk ("%s: Sending message\n", __FUNCTION__);	    
			}
			retval = ca_send_message (state, p_ca_message, arg);
			message_display(retval);

			break;

		case CA_RESET:
			if (dst_ca_verbose) {
				dprintk ("%s: Resetting DST\n", __FUNCTION__);	    
			}
			retval = dst_reset8820(state);
			message_display(retval);
			
			break;
	
		case CA_GET_SLOT_INFO:
			if (dst_ca_verbose) {
				dprintk ("%s: Getting Slot info\n", __FUNCTION__);	    
			}
			retval = ca_get_slot_info(state, p_ca_slot_info, arg);
			message_display(retval);
			
			break;

		case CA_GET_CAP:
			if (dst_ca_verbose) {
				dprintk ("%s: Getting Slot capabilities\n", __FUNCTION__);	    
			}
			retval = ca_get_slot_caps(state, p_ca_caps, arg);
			message_display(retval);
	
			break;

		case CA_GET_DESCR_INFO:
			if (dst_ca_verbose) {
				dprintk ("%s: Getting descrambler description\n", __FUNCTION__);	    
			}
			retval = ca_get_slot_descr(state, p_ca_message, arg);
			message_display(retval);
	
			break;

		case CA_SET_DESCR:
			if (dst_ca_verbose) {
				dprintk ("%s: Setting descrambler\n", __FUNCTION__);	    
			}
			retval = ca_set_slot_descr();
			message_display(retval);

			break;
			

	 	case CA_SET_PID:
			if (dst_ca_verbose) {
				dprintk ("%s: Setting PID\n", __FUNCTION__);	    
			}
			retval = ca_set_pid();
			message_display(retval);
			    	
		default:
			return -EOPNOTSUPP;
		};
        
	return (0);

}

static int message_display (int retval)
{
	if (retval < 0) {
		dprintk ("Failed.\n");
	}
	else if (retval == 0) {
		dprintk ("Success.\n");
	}
	else {
		dprintk ("Should not be here.\n");
	}
	
	return (0);
}

static int ca_set_slot_descr (void)
{
	/*	We could make this more graceful ?	*/
	return -EOPNOTSUPP;
}

static int ca_set_pid (void)
{
	/*	We could make this more graceful ?	*/
	return -EOPNOTSUPP;
}

static int ca_get_slot_caps (struct dst_state *state, struct ca_caps *p_ca_caps, void *arg)
{
	int retval, i;
	u8 slot_cap[256];
	u8 slot_command[8] = {0x07, 0x40, 0x00, 0x00, 0x02, 0x00, 0x00, 0xb7};

	retval = dst_put_ci(state, slot_command, sizeof(slot_command), slot_cap, 1);
	message_display(retval);
	/*	Will implement the rest soon	*/
	
	dprintk ("%s: Slot cap = [%d]\n", __FUNCTION__, slot_cap[7]);
	dprintk ("===================================\n");
	for (i = 0; i < 8; i++) {
		dprintk (" %d", slot_cap[i]);
	}
	dprintk ("\n");

	p_ca_caps->slot_num = 1;
	p_ca_caps->slot_type = 1;
	p_ca_caps->descr_num = slot_cap[7];	
	p_ca_caps->descr_type = 1;	


	if (copy_to_user((struct ca_caps *)arg, p_ca_caps, sizeof(struct ca_caps))) {		
		return -EFAULT;	
	}

	return (0);
}


static int ca_get_slot_descr (struct dst_state *state, struct ca_msg *p_ca_message, void *arg)
{


	return -EOPNOTSUPP;
}


static int ca_get_slot_info (struct dst_state *state, struct ca_slot_info *p_ca_slot_info, void *arg)
{
	int retval, i;
	u8 slot_info[8];
	u8 slot_command[8] = {0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfb};

	retval = dst_put_ci(state, slot_command, sizeof(slot_command), slot_info, 1);
	message_display(retval);
	
	/*	Will implement the rest soon	*/
	
	dprintk ("%s: Slot info = [%d]\n", __FUNCTION__, slot_info[3]);
	dprintk ("===================================\n");
	for (i = 0; i < 8; i++) {
		dprintk (" %d", slot_info[i]);
	}
	dprintk ("\n");

	if (slot_info[3] & 0x80) {
		p_ca_slot_info->flags = CA_CI_MODULE_PRESENT;
		p_ca_slot_info->num = 1;
		p_ca_slot_info->type = CA_CI;
	}
	else if (slot_info[3] & 0x40) {
		p_ca_slot_info->flags = CA_CI_MODULE_READY;
		p_ca_slot_info->num = 1;
		p_ca_slot_info->type = CA_CI;
	}
	else {
		p_ca_slot_info->flags = 0;
	}

	if (copy_to_user((struct ca_slot_info *)arg, p_ca_slot_info, sizeof(struct ca_slot_info))) {		
		return -EFAULT;
	}

	return (0);
}

static int ca_get_message (struct dst_state *state, struct ca_msg *p_ca_message, void *arg)
{
	/*	CA Get should not be harder than this	*/
	if (copy_to_user ((void *)arg, p_ca_message, sizeof (struct ca_msg))) {
		return -EFAULT;
	}

	return (0);
}

static int ca_send_message (struct dst_state *state, struct ca_msg *p_ca_message, void *arg)
{
	int i = 0;
	int retval = 0;
	unsigned int ca_message_header_len;
    
	u32 command = 0;
	struct ca_msg *hw_buffer;

	if ((hw_buffer = (struct ca_msg *) kmalloc (sizeof (struct ca_msg), GFP_KERNEL)) == NULL) {
		printk ("%s: Memory allocation failure\n", __FUNCTION__);
		return -ENOMEM;
	}

	dprintk ("%s\n", __FUNCTION__);
	
	if (copy_from_user (p_ca_message, (void *)arg, sizeof (struct ca_msg))) {
		return -EFAULT;
	}

	if (p_ca_message->msg) {

		ca_message_header_len = p_ca_message->length;	/* Restore it back when you are done	*/
	
		/*	EN50221 tag	*/
		command = 0;
	
		for (i = 0; i < 3; i++) {
			command = command | p_ca_message->msg[i];
			if (i < 2) {
				command = command << 8;
			}
		}
		dprintk ("%s:Command=[0x%x]\n", __FUNCTION__, command);
	
		switch (command) {
			case CA_PMT:
				dprintk ("Command = SEND_CA_PMT\n");
				retval = ca_set_pmt(state, p_ca_message, hw_buffer, 0, 0);
		    		message_display(retval);
		    
				break;

			case CA_PMT_REPLY:
				dprintk ("Command = CA_PMT_REPLY\n");
				/*	Have to handle the 2 basic types of cards here	*/
				retval = dst_check_ca_pmt (state, p_ca_message, hw_buffer);
				message_display(retval);
				
//				retval = ca_set_pmt(state, p_ca_message, hw_buffer, 1, 0);
//				message_display(retval);
				/*	Certain boards do behave different ?		*/
//				retval = ca_set_pmt(state, p_ca_message, hw_buffer, 1, 1);
//				message_display(retval);				

				break;
		}
	}
	return (0);
}
/*	Board supports CA PMT reply ?		*/
static int dst_check_ca_pmt (struct dst_state *state, struct ca_msg *p_ca_message, struct ca_msg *hw_buffer)
{
	int ca_pmt_reply_test = 0;
	int retval = 0;
	
	/*	Do test board			*/
	/*	Not there yet but soon		*/
	
	
	/*	CA PMT Reply capable		*/
	if (ca_pmt_reply_test) {
		retval = ca_set_pmt (state, p_ca_message, hw_buffer, 1, 1);
		message_display (retval);
	/*	Process CA PMT Reply		*/
	/*	will implement soon		*/
		dprintk ("Not there yet\n");
	
	}	
	/*	CA PMT Reply not capable	*/
	if (!ca_pmt_reply_test) {
		retval = ca_set_pmt (state, p_ca_message, hw_buffer, 0, 0);
		message_display (retval);
	/*	put a dummy message		*/

	}
	

	return (0);
}


static int ca_set_pmt (struct dst_state *state, struct ca_msg *p_ca_message, struct ca_msg *hw_buffer, u8 reply, u8 query)
{
	int i, retval = 0;
	int ca_desc_found = 0;

	unsigned int words, tag_len, hw_offset, buf_offset, prog_desc_len, temp_length;
	u32 length = 0;

	tag_len = 3;
    
	if (session) {
		hw_buffer->msg[2] = p_ca_message->msg[tag_len - 2];	/*	MSB		*/
		hw_buffer->msg[3] = p_ca_message->msg[tag_len - 3]; 	/*	LSB		*/
	}
	else {
		hw_buffer->msg[2] = 0x03;
		hw_buffer->msg[3] = 0x00;
	}

	/* Handle the length field (variable) */
	if (!(p_ca_message->msg[3] & 0x80)) {			/*	Length = 1		*/
		length = p_ca_message->msg[3] & 0x7f;
		words = 1;
	}
	else if (p_ca_message->msg[3] & 0x80) {			/*	Length = words		*/
		words = p_ca_message->msg[3] & 0x7f;
		for (i = 0; i < words; i++) {
			length = length | p_ca_message->msg[3 + i];
			length = length << 8;
		}
	}

	dprintk ("In string = [");
	for (i = 0; i < length; i++) {
		dprintk (" 0x%02x", p_ca_message->msg[i]);
	}
	dprintk ("]\n");


	if ( !(length > 248)) {					/*	Length = 255 - 7	*/
		length = (length - (tag_len + words)) + 7;
		hw_buffer->msg[0] = length;
		hw_buffer->msg[1] = 64;
		hw_buffer->msg[4] = 3;
		hw_buffer->msg[5] = hw_buffer->msg[0] - 7;
		hw_buffer->msg[6] = 0;
	}
	else {
		printk ("%s: Hardware Limitation, cannot send more than 248 words\n", __FUNCTION__);
	}

	hw_offset = 7;
	buf_offset = 4;
    
	/*	Copy from buffer to buffer	*/
	for (i = 0; i < p_ca_message->msg[3]; i++) {
		hw_buffer->msg[hw_offset + i] = p_ca_message->msg[buf_offset + i];
	}
	/*	Make the necessary changes before writing	*/
	hw_buffer->msg[11] = hw_buffer->msg[11] & 0x0f;		/*	prog info length	*/	
	hw_buffer->msg[12] = hw_buffer->msg[12] + 1;		/*	prog info length	*/	
	hw_buffer->length = hw_buffer->msg[0];			/*	length			*/
    
	prog_desc_len = 0;    
	prog_desc_len = ((prog_desc_len | hw_buffer->msg[11]) << 8) | hw_buffer->msg[12];
    
	dprintk ("Prog desc length =[%d]\n", prog_desc_len);

	if (prog_desc_len) {
		if (! reply) {
			hw_buffer->msg[13] = 0x01;	/*	CA PMT cmd id (OK descrambling)	*/
			dprintk ("CA PMT Command = Ok descrambling\n");
		}
		else {
			hw_buffer->msg[13] = 0x01;	/*	CA PMT cmd id (OK MMI)		*/
			dprintk ("CA PMT Command = OK MMI\n");
		}
		if (query) {
			hw_buffer->msg[13] = 0x03;	/*	CA PMT cmd id (query)		*/
			dprintk (" CA PMT Command = CA PMT query\n");
		}
		
		for (i = 14; i < (prog_desc_len + 13); i++) {
			if (hw_buffer->msg[i] == 0x09) {
				ca_desc_found = 1;
				dprintk ("CA descriptor found in program level.\n");
			}
		}
	}
	
	temp_length = prog_desc_len + 14;		/*	end programs, start streams	*/
	if (ca_desc_found) {				/*	CA descriptor at program level	*/
		for (i = temp_length; i < hw_buffer->length; i++) {
			hw_buffer->msg[i + 3] = 0x0;
			hw_buffer->msg[i + 4] = 0x0;
			i = i + 4;
		}	
	}
	else {						/*	CA descriptor at stream level	*/
							/*	Everything remains same ?	*/
	}
	
	do_checksum (hw_buffer);	
	retval = dst_put_ci (state, hw_buffer->msg, hw_buffer->length, hw_buffer->msg, reply);

	if (retval < 0) {
		dprintk ("%s: DST-CI Command failed.\n", __FUNCTION__);
		dprintk ("%s: Resetting DST.\n", __FUNCTION__);
		dst_reset8820(state); 
		return retval; 
	}
	else if (retval == 0) {
		dprintk ("%s: DST-CI Command succes.\n", __FUNCTION__);
	}
	else {
		dprintk ("%s: Should not be here.\n", __FUNCTION__);
		return (-1);
	}
    
	return (0);    
}


static int do_checksum (struct ca_msg *hw_buffer)
{
	int i;
	u8 checksum = 0;
    
	for (i = 0; i < (hw_buffer->length - 1); i++) {
		checksum = checksum + hw_buffer->msg[i];
	}
	hw_buffer->msg[hw_buffer->length] = ~hw_buffer->msg[hw_buffer->length] + 1;
	checksum = ~checksum + 1;    
	hw_buffer->msg[hw_buffer->length] = checksum;
	hw_buffer->length++;
	
	return (0);
}

static int dst_put_ci(struct dst_state *state, u8 *data, int len, u8 *ca_string, int read)
{
	int retval = 0;
	int length = 0;
	int dst_ack = 0;

	dst_i2c_enable(state);
	dst_reset8820(state);

	if (dst_ca_verbose)		
	    dprintk("%s:Write DST length=[%d]:\n",__FUNCTION__, len);

	retval = write_dst (state, data, len);
	if (retval < 0) {
		dst_i2c_disable(state);
		dprintk("%s: write not successful, maybe no card?\n", __FUNCTION__);
		return retval;
	}
	msleep(5);
	retval = read_dst (state, ca_string, 1);
	if (retval < 0) {
		dprintk("%s: read not successful\n", __FUNCTION__);
	}
	if (retval == 0) {
		dprintk("%s: read successful\n", __FUNCTION__);
		dst_ack = ca_string[0];
	}

	if (dst_ca_verbose) {
		dprintk("%s: ACK = [%d]\n", __FUNCTION__, dst_ack);
	}
	dst_i2c_disable(state);

	if (retval < 0) {
		dprintk("%s: read failure, maybe no Card/CAM ?\n", __FUNCTION__);
		return retval;
	}
	
	if (dst_ack != 0xff) {
		dprintk("%s: write not acknowledged.\n", __FUNCTION__);
		return retval;
	}

	if (read & dst_ack) {		
		retval = dst_wait_dst_ready(state);

		if (retval < 0) {
			dprintk ("%s: DST not ready.\n", __FUNCTION__);
			dst_i2c_disable(state);
			return retval;
		}

		retval = read_dst (state, ca_string, 1);
		if (retval < 0) {
			dprintk("%s: read not successful\n", __FUNCTION__);
		}
		if (retval == 0) {
			dprintk("%s: read successful\n", __FUNCTION__);
			length = ca_string[0];
		}
		
		dprintk ("%s: Read Length = [%d]\n", __FUNCTION__, length);

		if (length) {
			retval = read_dst (state, ca_string, (length - 1));
			if (retval < 0) {
				dprintk("%s: read not successful\n", __FUNCTION__);
			}
			if (retval == 0) {
				dprintk("%s: read successful\n", __FUNCTION__);
			}
		}
		dst_i2c_disable(state);
	
		if (retval < 0) {
			printk ("%s: Read failure.\n",__FUNCTION__);
			return (retval);
		}
	}
	return (0);
}


static int dst_ca_open (struct inode *inode, struct file *file)
{
	dprintk ("%s: Device opened [%p]\n", __FUNCTION__, file);    
	try_module_get(THIS_MODULE);

	return (0);
}

static int dst_ca_release (struct inode *inode, struct file *file)
{
	dprintk ("%s: Device closed.\n", __FUNCTION__);
	module_put(THIS_MODULE);

	return (0); 
}

static int dst_ca_read (struct file *file, char __user * buffer, size_t length, loff_t * offset)
{
	int bytes_read = 0;
	dprintk ("%s: Device read.\n", __FUNCTION__);

	return (bytes_read);
}

static int dst_ca_write (struct file *file, const char __user * buffer, size_t length, loff_t * offset)
{
	dprintk ("%s: Device write.\n", __FUNCTION__);
     
	return (0);
}


int dst_ca_attach(struct dst_state *dst, struct dvb_adapter *dvb_adapter)
{

	struct dvb_device *dvbdev;
        
	dprintk ("%s: registering DST-CA device\n", __FUNCTION__);
	dvb_register_device (dvb_adapter, &dvbdev, &dvbdev_ca, dst, DVB_DEVICE_CA);
	return (0);
}

EXPORT_SYMBOL(dst_ca_attach);

MODULE_DESCRIPTION("DST DVB-S/T/C Combo CA driver");
MODULE_AUTHOR("Manu Abraham");
MODULE_LICENSE("GPL");
